Package: qgarch
Type: Package
Title: Quadratic GARCH-in-Mean Models for Volatility Feedback
Version: 0.1.0
Authors@R: c(
    person(
        given = "Jedrzej",
        family = "Bialkowski",
        email = "jedrzej.bialkowski@canterbury.ac.nz",
        role = "aut"
    ),
    person(
        given = "Sanghyun",
        family = "Hong",
        email = "sanghyun.hong@canterbury.ac.nz",
        role = c("aut", "cre")
    ),
    person(
        given = "Moritz",
        family = "Wagner",
        email = "moritz.wagner@canterbury.ac.nz",
        role = "aut"
    )
  )
Description: Fits quadratic generalized autoregressive conditional
    heteroskedasticity-in-mean (QGARCH-M) models motivated by Campbell and
    Hentschel (1992). The package supports models with lambda fixed at zero,
    lambda restricted to a function of the remaining parameters, lambda
    estimated freely, and a threshold extension with state-dependent lambda.
    It also provides tools for starting values, estimation, forecasting,
    likelihood-ratio testing, moment diagnostics, and replication with the
    included monthly U.S. stock market dataset.
License: MIT + file LICENSE
Encoding: UTF-8
Depends: R (>= 4.1.0)
Imports: MASS
Suggests: testthat (>= 3.0.0)
LazyData: true
URL: https://github.com/sho-125/qgarch
BugReports: https://github.com/sho-125/qgarch/issues
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-05 03:06:02 UTC; sho125
Author: Jedrzej Bialkowski [aut],
  Sanghyun Hong [aut, cre],
  Moritz Wagner [aut]
Maintainer: Sanghyun Hong <sanghyun.hong@canterbury.ac.nz>
Repository: CRAN
Date/Publication: 2026-05-07 16:41:03 UTC
Built: R 4.5.2; ; 2026-05-07 19:30:50 UTC; unix

Package: qgarch
Type: Package
Title: Quadratic GARCH-in-Mean Models for Volatility Feedback
Version: 0.1.0
Authors@R: c(
    person(
        given = "Jedrzej",
        family = "Bialkowski",
        email = "jedrzej.bialkowski@canterbury.ac.nz",
        role = "aut"
    ),
    person(
        given = "Sanghyun",
        family = "Hong",
        email = "sanghyun.hong@canterbury.ac.nz",
        role = c("aut", "cre")
    ),
    person(
        given = "Moritz",
        family = "Wagner",
        email = "moritz.wagner@canterbury.ac.nz",
        role = "aut"
    )
  )
Description: Fits quadratic generalized autoregressive conditional
    heteroskedasticity-in-mean (QGARCH-M) models motivated by Campbell and
    Hentschel (1992). The package supports models with lambda fixed at zero,
    lambda restricted to a function of the remaining parameters, lambda
    estimated freely, and a threshold extension with state-dependent lambda.
    It also provides tools for starting values, estimation, forecasting,
    likelihood-ratio testing, moment diagnostics, and replication with the
    included monthly U.S. stock market dataset.
License: MIT + file LICENSE
Encoding: UTF-8
Depends: R (>= 4.1.0)
Imports: MASS
Suggests: testthat (>= 3.0.0)
LazyData: true
URL: https://github.com/sho-125/qgarch
BugReports: https://github.com/sho-125/qgarch/issues
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-05 03:06:02 UTC; sho125
Author: Jedrzej Bialkowski [aut],
  Sanghyun Hong [aut, cre],
  Moritz Wagner [aut]
Maintainer: Sanghyun Hong <sanghyun.hong@canterbury.ac.nz>
Repository: CRAN
Date/Publication: 2026-05-07 16:41:03 UTC
Built: R 4.6.0; ; 2026-05-07 17:08:06 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(coef,qgarch)
S3method(fitted,qgarch)
S3method(logLik,qgarch)
S3method(plot,qgarch)
S3method(predict,qgarch)
S3method(print,qgarch)
S3method(print,summary.qgarch)
S3method(residuals,qgarch)
S3method(summary,qgarch)
S3method(vcov,qgarch)
export(qgarch_default_starts)
export(qgarch_fit)
export(qgarch_lr_test)
export(qgarch_moments)

# qgarch

`qgarch` is an R package for estimating and working with quadratic
GARCH-in-mean models in the Campbell and Hentschel (1992)
volatility-feedback framework. 

We replicated and extended the original paper in: 
Jedrzej Bialkowski, Sanghyun Hong and Moritz Wagner (2025).
"Is no news still good news? Volatility feedback revisited." Pacific-Basin Finance Journal, Volume 91.
https://doi.org/10.1016/j.pacfin.2025.102708.

In the Campbell and Hentschel model, expected returns vary with
conditional variance, and stock returns include a volatility-feedback
term that helps capture asymmetric return dynamics. The framework is
designed to accommodate features such as predictive asymmetry, negative
skewness, and excess kurtosis in stock returns. See, John Y. Campbell, Ludger Hentschel (1992).
"No news is good news: An asymmetric model of changing volatility in stock returns."
Journal of Financial Economics, Volume 31, Issue 3, 281-318.
https://doi.org/10.1016/0304-405X(92)90037-X. 

## What the package does

`qgarch` provides tools to:

- fit generalized QGARCH-in-mean models with nonlinear maximum likelihood,
- work with several model specifications through a common interface,
- extract coefficients, fitted values, residuals, variance-covariance matrices,
  and log-likelihood values,
- generate forecasts of the conditional mean and conditional variance,
- compute moment diagnostics, and
- compare nested specifications with likelihood-ratio tests.

The main estimation function is `qgarch_fit()`. It supports four model
specifications:

- `model = "zero"`: sets `lambda = 0`,
- `model = "restricted"`: links `lambda` to the remaining parameters through
  the restricted mapping,
- `model = "free"`: estimates `lambda` freely, and
- `model = "threshold"`: allows a state-dependent `lambda_t`.

## Installation

You can install the development version of `qgarch` directly from GitHub:

```r
install.packages("remotes")
remotes::install_github("sho-125/qgarch")
library(qgarch)
```

## Example

The example below installs the package from GitHub, loads the bundled
monthly U.S. dataset, fits a QGARCH(1,1) model with freely estimated
volatility-feedback parameter `lambda`, and then plots the fitted
conditional variance and standardized residuals.

```r
# Install from GitHub
install.packages("remotes")
remotes::install_github("sho-125/qgarch")

# Load package
library(qgarch)

# Load bundled example data
data("us_monthly", package = "qgarch")

# Extract excess returns
x <- us_monthly$ER

# Fit a QGARCH(1,1) model
fit11 <- qgarch_fit(
  x,
  model = "free",
  arch_order = 1L,
  garch_order = 1L
)

# Review estimation results
print(fit11)
summary(fit11)
coef(fit11)
logLik(fit11)

# Plot fitted conditional variance
plot(fit11, which = "sigma2")

# Plot standardized residuals
plot(fit11, which = "standardized")

#' Compute sample moments for a series
#'
#' Computes the sample mean, variance, skewness, and excess kurtosis
#' for a numeric series, along with simple standard errors.
#'
#' @param x A numeric vector or a one-column data frame containing the series.
#' @param scale_mean_variance A finite scalar used to rescale the reported mean
#'   and variance and their standard errors.
#'
#' @return A list with elements:
#' \describe{
#'   \item{estimate}{Named numeric vector of moment estimates.}
#'   \item{std_error}{Named numeric vector of standard errors.}
#'   \item{n}{Number of usable observations.}
#' }
#' The returned object has class `"qgarch_moments"`.
#'
#' @export
qgarch_moments <- function(x, scale_mean_variance = 1) {
  if (!is.numeric(scale_mean_variance) ||
      length(scale_mean_variance) != 1L ||
      !is.finite(scale_mean_variance)) {
    stop("'scale_mean_variance' must be a single finite numeric value.", call. = FALSE)
  }
  
  x <- .validate_input_series(x)
  nn <- length(x)
  vx <- stats::var(x)
  sx <- stats::sd(x)
  
  out <- list(
    estimate = c(
      mean = mean(x) * scale_mean_variance,
      variance = vx * scale_mean_variance,
      skewness = .skewness(x),
      excess_kurtosis = .excess_kurtosis(x)
    ),
    std_error = c(
      mean = sx / sqrt(nn) * scale_mean_variance,
      variance = sqrt(2 * vx^2 / (nn - 1)) * scale_mean_variance,
      skewness = sqrt(6 * nn * (nn - 1) / ((nn - 2) * (nn + 1) * (nn + 3))),
      excess_kurtosis =
        2 * sqrt(6 * nn * (nn - 1) / ((nn - 2) * (nn + 1) * (nn + 3))) *
        sqrt((nn^2 - 1) / ((nn - 3) * (nn + 5)))
    ),
    n = nn
  )
  
  class(out) <- "qgarch_moments"
  out
}

#' Likelihood ratio test for nested qgarch models
#'
#' Compares two fitted `qgarch` models using the likelihood ratio test.
#'
#' @param restricted A fitted `qgarch` object representing the restricted model.
#' @param unrestricted A fitted `qgarch` object representing the unrestricted model.
#' @param df Degrees of freedom for the test. If `NULL`, it is computed as the
#'   difference in the number of estimated coefficients.
#'
#' @return A data frame with the likelihood ratio statistic, degrees of freedom,
#'   p-value, and the model names. The returned object has class
#'   `c("qgarch_lr_test", "data.frame")`.
#'
#' @export
qgarch_lr_test <- function(restricted, unrestricted, df = NULL) {
  if (!inherits(restricted, "qgarch") || !inherits(unrestricted, "qgarch")) {
    stop("Both inputs must be 'qgarch' objects.", call. = FALSE)
  }
  
  if (!is.numeric(restricted$nll) || length(restricted$nll) != 1L || !is.finite(restricted$nll)) {
    stop("The restricted model does not contain a finite negative log-likelihood.", call. = FALSE)
  }
  if (!is.numeric(unrestricted$nll) || length(unrestricted$nll) != 1L || !is.finite(unrestricted$nll)) {
    stop("The unrestricted model does not contain a finite negative log-likelihood.", call. = FALSE)
  }
  
  if (is.null(df)) {
    df <- length(unrestricted$coefficients) - length(restricted$coefficients)
  }
  
  if (!is.numeric(df) || length(df) != 1L || !is.finite(df) || df <= 0) {
    stop("'df' must be a single positive number.", call. = FALSE)
  }
  
  stat <- 2 * (restricted$nll - unrestricted$nll)
  
  if (!is.finite(stat)) {
    stop("Likelihood-ratio statistic is not finite.", call. = FALSE)
  }
  
  if (stat < 0) {
    warning(
      "Negative LR statistic encountered; setting statistic to 0. ",
      "This usually indicates numerical optimization error or that the ",
      "unrestricted fit did not improve on the restricted fit.",
      call. = FALSE
    )
    stat <- 0
  }
  
  pval <- stats::pchisq(stat, df = df, lower.tail = FALSE)
  
  out <- data.frame(
    statistic = stat,
    df = df,
    p.value = pval,
    restricted_model = restricted$model,
    unrestricted_model = unrestricted$model,
    row.names = NULL
  )
  
  class(out) <- c("qgarch_lr_test", class(out))
  out
}

test_that("moment diagnostics return named statistics", {
  data("us_monthly", package = "qgarch", envir = environment())
  x <- us_monthly$ER
  
  out <- qgarch_moments(x)
  
  expect_s3_class(out, "qgarch_moments")
  expect_true(is.list(out))
  expect_true(all(c("mean", "variance", "skewness", "excess_kurtosis") %in% names(out$estimate)))
  expect_true(all(c("mean", "variance", "skewness", "excess_kurtosis") %in% names(out$std_error)))
  expect_equal(out$n, length(x))
})

test_that("default starts return a non-empty list with named parameters", {
  data("us_monthly", package = "qgarch", envir = environment())
  x <- us_monthly$ER
  
  starts <- qgarch_default_starts(x, model = "free")
  
  expect_true(is.list(starts))
  expect_true(length(starts) >= 1)
  expect_true(is.numeric(starts[[1]]))
  expect_true(all(c("omega", "alpha1", "b", "beta1", "mu", "gamma", "lambda") %in% names(starts[[1]])))
})

test_that("qgarch_fit returns a usable qgarch object for zero model", {
  data("us_monthly", package = "qgarch", envir = environment())
  x <- us_monthly$ER
  
  fit <- qgarch_fit(x, model = "zero")
  
  expect_s3_class(fit, "qgarch")
  expect_equal(fit$model, "zero")
  expect_true(is.numeric(fit$loglik))
  expect_true(is.numeric(fit$nll))
  expect_equal(length(fit$x), length(x))
  expect_equal(length(fit$fitted), length(x))
  expect_equal(length(fit$residuals), length(x))
  expect_equal(length(fit$standardized_residuals), length(x))
  expect_equal(length(fit$sigma2), length(x))
  expect_true(all(c("omega", "alpha1", "b", "beta1", "mu", "gamma") %in% names(fit$coefficients)))
})

test_that("basic S3 methods work for qgarch objects", {
  data("us_monthly", package = "qgarch", envir = environment())
  x <- us_monthly$ER
  
  fit <- qgarch_fit(x, model = "zero")
  
  expect_true(is.numeric(coef(fit)))
  expect_true(is.matrix(vcov(fit)) || all(is.na(vcov(fit))))
  expect_true(is.numeric(fitted(fit)))
  expect_true(is.numeric(residuals(fit)))
  expect_true(is.numeric(residuals(fit, type = "standardized")))
  expect_s3_class(logLik(fit), "logLik")
  expect_s3_class(summary(fit), "summary.qgarch")
  
  fcst <- predict(fit, n.ahead = 3)
  expect_true(is.data.frame(fcst))
  expect_equal(nrow(fcst), 3)
  expect_true(all(c("step", "mean", "sigma2", "sigma") %in% names(fcst)))
})

test_that("likelihood ratio test returns a data frame with expected columns", {
  data("us_monthly", package = "qgarch", envir = environment())
  x <- us_monthly$ER
  
  fit_restricted <- qgarch_fit(x, model = "zero")
  fit_unrestricted <- qgarch_fit(x, model = "free")
  
  lr <- suppressWarnings(qgarch_lr_test(fit_restricted, fit_unrestricted))
  
  expect_s3_class(lr, "qgarch_lr_test")
  expect_true(is.data.frame(lr))
  expect_true(all(c("statistic", "df", "p.value", "restricted_model", "unrestricted_model") %in% names(lr)))
  expect_equal(nrow(lr), 1)
})

`%||%` <- function(x, y) if (is.null(x)) y else x

.get_alpha_beta <- function(object) {
  co <- object$coefficients
  arch_order <- object$arch_order %||% 1L
  garch_order <- object$garch_order %||% 1L
  
  alpha_names <- paste0("alpha", seq_len(arch_order))
  beta_names <- paste0("beta", seq_len(garch_order))
  
  if (all(alpha_names %in% names(co))) {
    alpha <- unname(co[alpha_names])
    names(alpha) <- alpha_names
  } else if ("alpha" %in% names(co) && arch_order == 1L) {
    alpha <- unname(co["alpha"])
    names(alpha) <- "alpha1"
  } else {
    stop("Could not find ARCH coefficients in object.", call. = FALSE)
  }
  
  if (all(beta_names %in% names(co))) {
    beta <- unname(co[beta_names])
    names(beta) <- beta_names
  } else if ("beta" %in% names(co) && garch_order == 1L) {
    beta <- unname(co["beta"])
    names(beta) <- "beta1"
  } else {
    stop("Could not find GARCH coefficients in object.", call. = FALSE)
  }
  
  list(
    alpha = alpha,
    beta = beta,
    arch_order = arch_order,
    garch_order = garch_order
  )
}

#' Extract coefficients from a qgarch model
#'
#' @param object A fitted `qgarch` object.
#' @param ... Additional arguments. Supports `type = "estimated"` or
#'   `type = "full"`.
#'
#' @return A named numeric vector of coefficients.
#' @export
coef.qgarch <- function(object, ...) {
  dots <- list(...)
  type <- dots$type %||% "estimated"
  type <- match.arg(type, c("estimated", "full"))
  
  if (type == "estimated") object$coefficients else object$coefficients_full
}

#' Variance-covariance matrix for a qgarch model
#'
#' @param object A fitted `qgarch` object.
#' @param ... Unused.
#'
#' @return A variance-covariance matrix.
#' @export
vcov.qgarch <- function(object, ...) {
  object$vcov
}

#' Fitted values from a qgarch model
#'
#' @param object A fitted `qgarch` object.
#' @param ... Unused.
#'
#' @return A numeric vector of fitted values.
#' @export
fitted.qgarch <- function(object, ...) {
  object$fitted
}

#' Residuals from a qgarch model
#'
#' @param object A fitted `qgarch` object.
#' @param type Type of residuals to return: `"raw"`, `"standardized"`,
#'   or `"eta"`.
#' @param ... Unused.
#'
#' @return A numeric vector of residuals.
#' @export
residuals.qgarch <- function(object, type = c("raw", "standardized", "eta"), ...) {
  type <- match.arg(type)
  
  switch(
    type,
    raw = object$residuals,
    standardized = object$standardized_residuals,
    eta = object$eta
  )
}

#' Log-likelihood for a qgarch model
#'
#' @param object A fitted `qgarch` object.
#' @param ... Unused.
#'
#' @return An object of class `"logLik"`.
#' @export
logLik.qgarch <- function(object, ...) {
  out <- object$loglik
  attr(out, "df") <- length(object$coefficients)
  attr(out, "nobs") <- length(object$x)
  class(out) <- "logLik"
  out
}

#' Forecast from a generalized qgarch(m, n) model
#'
#' @param object A fitted `qgarch` object.
#' @param n.ahead Number of periods ahead to forecast.
#' @param ... Unused.
#'
#' @return A data frame with forecast horizon, conditional mean,
#'   conditional variance, and conditional standard deviation.
#' @export
predict.qgarch <- function(object, n.ahead = 1L, ...) {
  if (!is.numeric(n.ahead) || length(n.ahead) != 1L || !is.finite(n.ahead)) {
    stop("'n.ahead' must be a single finite number.", call. = FALSE)
  }
  
  n.ahead <- as.integer(n.ahead)
  if (n.ahead < 1L) {
    stop("'n.ahead' must be at least 1.", call. = FALSE)
  }
  
  blocks <- .get_alpha_beta(object)
  alpha <- blocks$alpha
  beta <- blocks$beta
  arch_order <- blocks$arch_order
  garch_order <- blocks$garch_order
  
  co <- object$coefficients
  omega <- unname(co["omega"])
  b <- unname(co["b"])
  mu <- unname(co["mu"])
  gamma <- unname(co["gamma"])
  
  eta_hist <- object$eta
  sigma_hist <- object$sigma2
  uncond <- object$unconditional_variance %||% stats::var(object$x, na.rm = TRUE)
  
  sigma_fcst <- numeric(n.ahead)
  
  for (h in seq_len(n.ahead)) {
    arch_term <- 0
    for (j in seq_len(arch_order)) {
      idx <- h - j
      if (idx <= 0L) {
        hist_pos <- length(eta_hist) + idx
        eta_lag <- if (hist_pos >= 1L) eta_hist[hist_pos] else 0
        arch_piece <- (eta_lag - b)^2
      } else {
        arch_piece <- sigma_fcst[idx] + b^2
      }
      arch_term <- arch_term + alpha[j] * arch_piece
    }
    
    garch_term <- 0
    for (k in seq_len(garch_order)) {
      idx <- h - k
      if (idx <= 0L) {
        hist_pos <- length(sigma_hist) + idx
        sigma_lag <- if (hist_pos >= 1L) sigma_hist[hist_pos] else uncond
      } else {
        sigma_lag <- sigma_fcst[idx]
      }
      garch_term <- garch_term + beta[k] * sigma_lag
    }
    
    sigma_fcst[h] <- omega + arch_term + garch_term
  }
  
  data.frame(
    step = seq_len(n.ahead),
    mean = mu + gamma * sigma_fcst,
    sigma2 = sigma_fcst,
    sigma = sqrt(sigma_fcst),
    row.names = NULL
  )
}

#' Summarize a qgarch model
#'
#' @param object A fitted `qgarch` object.
#' @param ... Unused.
#'
#' @return An object of class `"summary.qgarch"`.
#' @export
summary.qgarch <- function(object, ...) {
  est <- object$coefficients
  se <- object$standard_errors
  
  if (is.null(se) || length(se) != length(est)) {
    se <- rep(NA_real_, length(est))
    names(se) <- names(est)
  }
  
  z <- rep(NA_real_, length(est))
  ok <- is.finite(est) & is.finite(se) & se > 0
  z[ok] <- est[ok] / se[ok]
  
  p <- rep(NA_real_, length(est))
  p[ok] <- 2 * stats::pnorm(abs(z[ok]), lower.tail = FALSE)
  
  coef_table <- cbind(
    Estimate = est,
    Std.Error = se,
    z.value = z,
    p.value = p
  )
  
  if (!is.null(object$implied_lambda)) {
    lambda_z <- NA_real_
    lambda_p <- NA_real_
    
    if (is.finite(object$implied_lambda_se) && object$implied_lambda_se > 0) {
      lambda_z <- object$implied_lambda / object$implied_lambda_se
      lambda_p <- 2 * stats::pnorm(abs(lambda_z), lower.tail = FALSE)
    }
    
    coef_table <- rbind(
      coef_table,
      lambda = c(
        Estimate = object$implied_lambda,
        Std.Error = object$implied_lambda_se,
        z.value = lambda_z,
        p.value = lambda_p
      )
    )
  }
  
  structure(
    list(
      model = object$model,
      arch_order = object$arch_order %||% 1L,
      garch_order = object$garch_order %||% 1L,
      coefficients = coef_table,
      loglik = object$loglik,
      nobs = length(object$x),
      convergence = object$convergence,
      unconditional_variance = object$unconditional_variance,
      rho = object$rho
    ),
    class = "summary.qgarch"
  )
}

#' Print a qgarch model
#'
#' @param x A fitted `qgarch` object.
#' @param digits Number of digits to print.
#' @param ... Unused.
#'
#' @return The input object, invisibly.
#' @export
print.qgarch <- function(x, digits = max(3L, getOption("digits") - 2L), ...) {
  arch_order <- x$arch_order %||% 1L
  garch_order <- x$garch_order %||% 1L
  
  cat("QGARCH fit\n")
  cat("  model:", x$model, "\n")
  cat("  order:", sprintf("(%d, %d)", arch_order, garch_order), "\n")
  cat("  observations:", length(x$x), "\n")
  cat("  log-likelihood:", format(round(x$loglik, digits), nsmall = digits), "\n")
  cat("  convergence code:", x$convergence, "\n\n")
  print(round(x$coefficients_full, digits))
  invisible(x)
}

#' Print a qgarch summary
#'
#' @param x An object of class `"summary.qgarch"`.
#' @param digits Number of digits to print.
#' @param ... Unused.
#'
#' @return The input object, invisibly.
#' @export
print.summary.qgarch <- function(x, digits = max(3L, getOption("digits") - 2L), ...) {
  cat("QGARCH summary\n")
  cat("  model:", x$model, "\n")
  cat("  order:", sprintf("(%d, %d)", x$arch_order, x$garch_order), "\n")
  cat("  observations:", x$nobs, "\n")
  cat("  log-likelihood:", format(round(x$loglik, digits), nsmall = digits), "\n")
  cat("  convergence code:", x$convergence, "\n")
  cat(
    "  unconditional variance:",
    format(round(x$unconditional_variance, digits), nsmall = digits),
    "\n"
  )
  if (identical(x$model, "restricted") && !is.null(x$rho)) {
    cat("  rho used in restricted mapping:", x$rho, "\n")
  }
  cat("\nCoefficients:\n")
  print(round(x$coefficients, digits))
  invisible(x)
}

#' Plot a qgarch model
#'
#' @param x A fitted `qgarch` object.
#' @param which Which plot to show: `"sigma2"` or `"standardized"`.
#' @param ... Additional graphical arguments passed to [graphics::plot()].
#'
#' @return The input object, invisibly.
#' @export
plot.qgarch <- function(x, which = c("sigma2", "standardized"), ...) {
  which <- match.arg(which)
  
  if (which == "sigma2") {
    graphics::plot(
      x$sigma2,
      type = "l",
      xlab = "Time",
      ylab = expression(sigma[t]^2),
      ...
    )
  } else {
    graphics::plot(
      x$standardized_residuals,
      type = "h",
      xlab = "Time",
      ylab = "Standardized residual",
      ...
    )
    graphics::abline(h = 0, lty = 2)
  }
  
  invisible(x)
}

#' Monthly U.S. data
#'
#' Monthly data used in examples and tests for the qgarch package.
#'
#' @format A data frame with monthly observations and 3 variables:
#' \describe{
#'   \item{YearMon}{Month of the observation.}
#'   \item{Mktrf}{Market excess return.}
#'   \item{ER}{Excess return series.}
#' }
#'
#' @source Prepared by the package author.
"us_monthly"

# Internal penalty used by likelihood functions.
.qgarch_big_penalty <- 1e200

.validate_input_series <- function(x) {
  if (is.data.frame(x) && ncol(x) == 1L) {
    x <- x[[1L]]
  }
  
  x <- as.numeric(x)
  
  if (length(x) < 10L) {
    stop("'x' must contain at least 10 observations.", call. = FALSE)
  }
  
  if (anyNA(x) || any(!is.finite(x))) {
    stop("'x' must contain only finite, non-missing values.", call. = FALSE)
  }
  
  x
}

.validate_orders <- function(arch_order, garch_order) {
  if (!is.numeric(arch_order) || length(arch_order) != 1L || !is.finite(arch_order)) {
    stop("'arch_order' must be a single positive integer.", call. = FALSE)
  }
  if (!is.numeric(garch_order) || length(garch_order) != 1L || !is.finite(garch_order)) {
    stop("'garch_order' must be a single positive integer.", call. = FALSE)
  }
  
  arch_order <- as.integer(arch_order)
  garch_order <- as.integer(garch_order)
  
  if (arch_order < 1L || garch_order < 1L) {
    stop("'arch_order' and 'garch_order' must both be at least 1.", call. = FALSE)
  }
  
  list(arch_order = arch_order, garch_order = garch_order)
}

.unconditional_variance <- function(omega, alpha, b, beta) {
  alpha_sum <- sum(alpha)
  beta_sum <- sum(beta)
  den <- 1 - alpha_sum - beta_sum
  
  if (!is.finite(den) || den <= 0 || !is.finite(omega) || omega <= 0) {
    return(NA_real_)
  }
  if (any(!is.finite(alpha)) || any(alpha < 0)) {
    return(NA_real_)
  }
  if (any(!is.finite(beta)) || any(beta < 0)) {
    return(NA_real_)
  }
  if (!is.finite(b)) {
    return(NA_real_)
  }
  
  (omega + alpha_sum * b^2) / den
}

.clean_sym_matrix <- function(mat) {
  if (is.null(mat) || !is.matrix(mat)) {
    return(NULL)
  }
  0.5 * (mat + t(mat))
}

.safe_invert <- function(mat, tol = 1e-7) {
  out <- tryCatch(
    solve(mat),
    error = function(e) NULL,
    warning = function(w) NULL
  )
  
  if (is.null(out)) {
    out <- MASS::ginv(mat, tol = tol)
  }
  
  out
}

.regularize_pd <- function(mat, rel_floor = 1e-8, abs_floor = 1e-10) {
  mat <- .clean_sym_matrix(mat)
  
  if (is.null(mat) || any(!is.finite(mat))) {
    return(NULL)
  }
  
  ee <- tryCatch(
    eigen(mat, symmetric = TRUE),
    error = function(e) NULL
  )
  if (is.null(ee)) {
    return(NULL)
  }
  
  max_eval <- max(abs(ee$values), na.rm = TRUE)
  floor_val <- max(max_eval * rel_floor, abs_floor)
  
  vals <- ee$values
  vals[!is.finite(vals)] <- floor_val
  vals[vals < floor_val] <- floor_val
  
  out <- ee$vectors %*% (vals * t(ee$vectors))
  .clean_sym_matrix(out)
}

.make_se_from_vcov <- function(vcov_mat, tol = 1e-12) {
  vcov_mat <- .clean_sym_matrix(vcov_mat)
  
  if (is.null(vcov_mat) || any(!is.finite(vcov_mat))) {
    return(list(se = NULL, vcov = NULL))
  }
  
  dv <- diag(vcov_mat)
  se <- rep(NA_real_, length(dv))
  ok <- is.finite(dv) & dv > tol
  se[ok] <- sqrt(dv[ok])
  
  list(se = se, vcov = vcov_mat)
}

.skewness <- function(x) {
  m <- mean(x)
  s <- stats::sd(x)
  
  if (!is.finite(s) || s <= 0) {
    return(NA_real_)
  }
  
  mean((x - m)^3) / s^3
}

.excess_kurtosis <- function(x) {
  m <- mean(x)
  s <- stats::sd(x)
  
  if (!is.finite(s) || s <= 0) {
    return(NA_real_)
  }
  
  mean((x - m)^4) / s^4 - 3
}

.lag_weights <- function(k, pattern = c("geom", "equal", "front"), decay = 0.65) {
  pattern <- match.arg(pattern)
  
  w <- switch(
    pattern,
    geom = decay^(seq_len(k) - 1L),
    equal = rep(1, k),
    front = rev(seq_len(k))
  )
  
  w / sum(w)
}

.parameter_names <- function(model, arch_order, garch_order) {
  model <- match.arg(model, c("zero", "restricted", "free", "threshold"))
  
  base <- c(
    "omega",
    paste0("alpha", seq_len(arch_order)),
    "b",
    paste0("beta", seq_len(garch_order)),
    "mu",
    "gamma"
  )
  
  extra <- switch(
    model,
    zero = character(0),
    restricted = character(0),
    free = "lambda",
    threshold = c("lambda1", "lambda2")
  )
  
  c(base, extra)
}

.make_named_start <- function(model, arch_order, garch_order, values) {
  nm <- .parameter_names(model, arch_order = arch_order, garch_order = garch_order)
  values <- as.numeric(values)
  
  if (length(values) != length(nm)) {
    stop(
      "Starting values have length ", length(values),
      " but model/order combination requires ", length(nm), ".",
      call. = FALSE
    )
  }
  
  stats::setNames(values, nm)
}

.target_omega <- function(target_var, alpha_total, b, beta_total) {
  out <- target_var * (1 - alpha_total - beta_total) - alpha_total * b^2
  max(out, 1e-8)
}

.compose_start_values <- function(model,
                                  arch_order,
                                  garch_order,
                                  target_var,
                                  alpha_total,
                                  b,
                                  beta_total,
                                  mu,
                                  gamma,
                                  lambda = NULL,
                                  lambda1 = NULL,
                                  lambda2 = NULL,
                                  alpha_pattern = c("geom", "equal", "front"),
                                  beta_pattern = c("geom", "equal", "front")) {
  alpha_pattern <- match.arg(alpha_pattern)
  beta_pattern <- match.arg(beta_pattern)
  
  alpha <- alpha_total * .lag_weights(arch_order, pattern = alpha_pattern)
  beta <- beta_total * .lag_weights(garch_order, pattern = beta_pattern)
  omega <- .target_omega(target_var, alpha_total = sum(alpha), b = b, beta_total = sum(beta))
  
  out <- c(
    omega,
    alpha,
    b,
    beta,
    mu,
    gamma
  )
  
  if (model == "free") {
    out <- c(out, lambda)
  } else if (model == "threshold") {
    out <- c(out, lambda1, lambda2)
  }
  
  .make_named_start(model, arch_order, garch_order, out)
}

.restricted_lambda_value <- function(alpha, beta, gamma, rho = 1) {
  alpha_sum <- sum(alpha)
  beta_sum <- sum(beta)
  den <- 1 - rho * (alpha_sum + beta_sum)
  
  if (!is.finite(den) || den <= 0) {
    return(NA_real_)
  }
  
  rho * gamma * alpha_sum / den
}

#' Default starting values for qgarch estimation
#'
#' Creates a compact but order-adaptive set of starting values for the selected
#' qgarch model.
#'
#' @param x A numeric vector or one-column data frame containing the observed
#'   series.
#' @param model Character string specifying the model variant. Must be one of
#'   `"zero"`, `"restricted"`, `"free"`, or `"threshold"`.
#' @param arch_order Positive integer giving the ARCH lag order.
#' @param garch_order Positive integer giving the GARCH lag order.
#'
#' @return A list of named numeric vectors containing candidate starting values.
#' @export
qgarch_default_starts <- function(x,
                                  model = c("zero", "restricted", "free", "threshold"),
                                  arch_order = 1L,
                                  garch_order = 1L) {
  model <- match.arg(model)
  orders <- .validate_orders(arch_order, garch_order)
  arch_order <- orders$arch_order
  garch_order <- orders$garch_order
  
  x <- .validate_input_series(x)
  
  vx <- stats::var(x)
  sx <- stats::sd(x)
  mx <- mean(x)
  b_small <- max(sx * 0.10, 1e-4)
  total_order <- arch_order + garch_order
  
  profiles <- list(
    list(alpha_total = 0.08, beta_total = 0.84, b = b_small,        mu = mx, gamma = 0.00),
    list(alpha_total = 0.10, beta_total = 0.82, b = 0.00,           mu = mx, gamma = 0.05),
    list(alpha_total = 0.06, beta_total = 0.88, b = 0.75 * b_small, mu = mx, gamma = 0.15),
    list(alpha_total = 0.12, beta_total = 0.75, b = 0.50 * b_small, mu = mx, gamma = -0.02)
  )
  
  if (total_order <= 4L) {
    pattern_pairs <- list(
      c("geom", "geom"),
      c("equal", "equal")
    )
  } else {
    pattern_pairs <- list(
      c("geom", "geom")
    )
  }
  
  starts <- list()
  
  for (prof in profiles) {
    for (pp in pattern_pairs) {
      alpha_pattern <- pp[1]
      beta_pattern <- pp[2]
      
      base <- .compose_start_values(
        model = "restricted",
        arch_order = arch_order,
        garch_order = garch_order,
        target_var = vx,
        alpha_total = prof$alpha_total,
        b = prof$b,
        beta_total = prof$beta_total,
        mu = prof$mu,
        gamma = prof$gamma,
        alpha_pattern = alpha_pattern,
        beta_pattern = beta_pattern
      )
      
      if (model %in% c("zero", "restricted")) {
        starts[[length(starts) + 1L]] <- .make_named_start(model, arch_order, garch_order, base)
        
      } else if (model == "free") {
        alpha_idx <- paste0("alpha", seq_len(arch_order))
        beta_idx <- paste0("beta", seq_len(garch_order))
        
        lam_restricted <- .restricted_lambda_value(
          alpha = base[alpha_idx],
          beta = base[beta_idx],
          gamma = base["gamma"],
          rho = 1
        )
        
        lambda_grid <- unique(c(0, lam_restricted, 0.25))
        lambda_grid <- lambda_grid[is.finite(lambda_grid)]
        
        for (lam in lambda_grid) {
          tmp <- c(base, lambda = lam)
          starts[[length(starts) + 1L]] <- .make_named_start(model, arch_order, garch_order, tmp)
        }
        
      } else if (model == "threshold") {
        lam1_grid <- c(0, 0.25)
        lam2_grid <- c(0, 0.10)
        
        for (lam1 in lam1_grid) {
          for (lam2 in lam2_grid) {
            tmp <- c(base, lambda1 = lam1, lambda2 = lam2)
            starts[[length(starts) + 1L]] <- .make_named_start(model, arch_order, garch_order, tmp)
          }
        }
      }
    }
  }
  
  dedup_key <- vapply(
    starts,
    function(z) paste(signif(unname(z), 8), collapse = "|"),
    character(1)
  )
  
  starts[!duplicated(dedup_key)]
}

.solve_eta_quadratic <- function(obs, sigma2, mu, gamma, lambda, kappa, tol = 1e-10) {
  if (abs(lambda) <= tol) {
    return((obs - mu - gamma * sigma2) / kappa)
  }
  
  rhs <- obs - mu - (gamma + lambda) * sigma2
  disc <- kappa^2 - 4 * lambda * rhs
  
  if (!is.finite(disc) || disc < 0) {
    return(NA_real_)
  }
  
  (kappa - sqrt(disc)) / (2 * lambda)
}

.qgarch_nll <- function(theta,
                        x,
                        model,
                        arch_order,
                        garch_order,
                        rho = 1,
                        indicator = NULL,
                        return_details = FALSE) {
  n <- length(x)
  eta <- numeric(n)
  sigma2 <- numeric(n)
  loglik_i <- numeric(n)
  
  idx <- 1L
  omega <- theta[idx]
  idx <- idx + 1L
  
  alpha <- theta[idx:(idx + arch_order - 1L)]
  idx <- idx + arch_order
  
  b <- theta[idx]
  idx <- idx + 1L
  
  beta <- theta[idx:(idx + garch_order - 1L)]
  idx <- idx + garch_order
  
  mu <- theta[idx]
  idx <- idx + 1L
  
  gamma <- theta[idx]
  idx <- idx + 1L
  
  if (!is.finite(omega) || omega <= 0) {
    return(if (return_details) NULL else .qgarch_big_penalty)
  }
  if (any(!is.finite(alpha)) || any(alpha < 0)) {
    return(if (return_details) NULL else .qgarch_big_penalty)
  }
  if (any(!is.finite(beta)) || any(beta < 0)) {
    return(if (return_details) NULL else .qgarch_big_penalty)
  }
  if (sum(alpha) + sum(beta) >= 0.9995) {
    return(if (return_details) NULL else .qgarch_big_penalty)
  }
  
  if (model == "zero") {
    lambda_vec <- rep(0, n)
  } else if (model == "restricted") {
    lam <- .restricted_lambda_value(alpha, beta, gamma, rho = rho)
    if (!is.finite(lam)) {
      return(if (return_details) NULL else .qgarch_big_penalty)
    }
    lambda_vec <- rep(lam, n)
  } else if (model == "free") {
    lam <- theta[idx]
    if (!is.finite(lam) || abs(lam) > 50) {
      return(if (return_details) NULL else .qgarch_big_penalty)
    }
    lambda_vec <- rep(lam, n)
  } else if (model == "threshold") {
    lambda1 <- theta[idx]
    lambda2 <- theta[idx + 1L]
    if (!is.finite(lambda1) || !is.finite(lambda2) || max(abs(c(lambda1, lambda2))) > 50) {
      return(if (return_details) NULL else .qgarch_big_penalty)
    }
    lambda_vec <- lambda1 + lambda2 * indicator
  } else {
    stop("Unknown model.", call. = FALSE)
  }
  
  kappa_vec <- 1 + 2 * lambda_vec * b
  
  uncond <- .unconditional_variance(omega, alpha, b, beta)
  if (!is.finite(uncond)) {
    return(if (return_details) NULL else .qgarch_big_penalty)
  }
  
  for (i in seq_len(n)) {
    arch_term <- 0
    for (j in seq_len(arch_order)) {
      eta_lag <- if (i - j > 0L) eta[i - j] else 0
      arch_term <- arch_term + alpha[j] * (eta_lag - b)^2
    }
    
    garch_term <- 0
    for (k in seq_len(garch_order)) {
      sigma_lag <- if (i - k > 0L) sigma2[i - k] else uncond
      garch_term <- garch_term + beta[k] * sigma_lag
    }
    
    sigma2[i] <- omega + arch_term + garch_term
    
    if (model == "zero") {
      eta[i] <- x[i] - mu - gamma * sigma2[i]
    } else {
      eta[i] <- .solve_eta_quadratic(
        obs = x[i],
        sigma2 = sigma2[i],
        mu = mu,
        gamma = gamma,
        lambda = lambda_vec[i],
        kappa = kappa_vec[i]
      )
    }
    
    if (!is.finite(sigma2[i]) || sigma2[i] <= 0 || !is.finite(eta[i])) {
      return(if (return_details) NULL else .qgarch_big_penalty)
    }
    
    if (model == "zero") {
      loglik_i[i] <- -0.5 * log(2 * pi * sigma2[i]) - 0.5 * (eta[i]^2 / sigma2[i])
    } else {
      jac <- kappa_vec[i] - 2 * lambda_vec[i] * eta[i]
      if (!is.finite(jac) || jac <= 0) {
        return(if (return_details) NULL else .qgarch_big_penalty)
      }
      loglik_i[i] <- -log(jac) - 0.5 * log(2 * pi * sigma2[i]) - 0.5 * (eta[i]^2 / sigma2[i])
    }
  }
  
  nll <- -sum(loglik_i)
  
  if (return_details) {
    return(list(
      nll = nll,
      eta = eta,
      sigma2 = sigma2,
      uncond_var = uncond,
      lambda = lambda_vec,
      kappa = kappa_vec,
      loglik_contrib = loglik_i
    ))
  }
  
  nll
}

.numeric_score_matrix <- function(theta,
                                  x,
                                  model,
                                  arch_order,
                                  garch_order,
                                  rho,
                                  indicator,
                                  eps = 1e-5) {
  p <- length(theta)
  base <- .qgarch_nll(
    theta,
    x = x,
    model = model,
    arch_order = arch_order,
    garch_order = garch_order,
    rho = rho,
    indicator = indicator,
    return_details = TRUE
  )
  
  if (is.null(base)) {
    return(NULL)
  }
  
  n <- length(base$loglik_contrib)
  score <- matrix(NA_real_, nrow = n, ncol = p)
  
  for (j in seq_len(p)) {
    step <- eps * max(1, abs(theta[j]))
    
    theta_up <- theta
    theta_dn <- theta
    theta_up[j] <- theta_up[j] + step
    theta_dn[j] <- theta_dn[j] - step
    
    up <- .qgarch_nll(
      theta_up,
      x = x,
      model = model,
      arch_order = arch_order,
      garch_order = garch_order,
      rho = rho,
      indicator = indicator,
      return_details = TRUE
    )
    dn <- .qgarch_nll(
      theta_dn,
      x = x,
      model = model,
      arch_order = arch_order,
      garch_order = garch_order,
      rho = rho,
      indicator = indicator,
      return_details = TRUE
    )
    
    if (is.null(up) || is.null(dn)) {
      return(NULL)
    }
    
    score[, j] <- (up$loglik_contrib - dn$loglik_contrib) / (2 * step)
  }
  
  colnames(score) <- names(theta)
  score
}

.compute_vcov <- function(theta,
                          raw_hessian,
                          x,
                          model,
                          arch_order,
                          garch_order,
                          rho,
                          indicator,
                          vcov_type = c("auto", "sandwich", "hessian", "opg", "none")) {
  vcov_type <- match.arg(vcov_type)
  est_names <- names(theta)
  
  empty <- list(
    vcov = matrix(
      NA_real_,
      nrow = length(theta),
      ncol = length(theta),
      dimnames = list(est_names, est_names)
    ),
    se = stats::setNames(rep(NA_real_, length(theta)), est_names),
    method = "none"
  )
  
  if (vcov_type == "none") {
    return(empty)
  }
  
  make_result <- function(vcov_mat, method) {
    se_info <- .make_se_from_vcov(vcov_mat)
    if (is.null(se_info$se) || !any(is.finite(se_info$se))) {
      return(NULL)
    }
    
    names(se_info$se) <- est_names
    rownames(se_info$vcov) <- est_names
    colnames(se_info$vcov) <- est_names
    
    list(vcov = se_info$vcov, se = se_info$se, method = method)
  }
  
  H_inv <- NULL
  
  if (vcov_type %in% c("auto", "sandwich", "hessian")) {
    H_raw <- .clean_sym_matrix(raw_hessian)
    
    H_use <- NULL
    if (!is.null(H_raw) && all(is.finite(H_raw))) {
      H_use <- H_raw
    } else {
      H_num <- tryCatch(
        stats::optimHess(
          par = unname(theta),
          fn = .qgarch_nll,
          x = x,
          model = model,
          arch_order = arch_order,
          garch_order = garch_order,
          rho = rho,
          indicator = indicator
        ),
        error = function(e) NULL,
        warning = function(w) NULL
      )
      H_num <- .clean_sym_matrix(H_num)
      
      if (!is.null(H_num) && all(is.finite(H_num))) {
        H_use <- H_num
      }
    }
    
    if (!is.null(H_use)) {
      H_reg <- .regularize_pd(H_use)
      H_inv <- .clean_sym_matrix(.safe_invert(H_reg))
    }
    
    if (!is.null(H_inv) && all(is.finite(H_inv))) {
      if (vcov_type == "hessian") {
        out <- make_result(H_inv, "hessian")
        return(if (is.null(out)) empty else out)
      }
      
      if (vcov_type == "auto") {
        out <- make_result(H_inv, "hessian")
        if (!is.null(out)) {
          return(out)
        }
      }
    }
  }
  
  if (vcov_type %in% c("auto", "sandwich", "opg")) {
    score_mat <- .numeric_score_matrix(
      theta = unname(theta),
      x = x,
      model = model,
      arch_order = arch_order,
      garch_order = garch_order,
      rho = rho,
      indicator = indicator
    )
    
    S <- NULL
    if (!is.null(score_mat) && all(is.finite(score_mat))) {
      S <- .regularize_pd(crossprod(score_mat))
    }
    
    if (vcov_type == "opg") {
      if (!is.null(S) && all(is.finite(S))) {
        S_inv <- .clean_sym_matrix(.safe_invert(S))
        out <- make_result(S_inv, "opg")
        return(if (is.null(out)) empty else out)
      }
      return(empty)
    }
    
    if (vcov_type == "sandwich") {
      if (!is.null(H_inv) && !is.null(S) && all(is.finite(H_inv)) && all(is.finite(S))) {
        out <- make_result(.clean_sym_matrix(H_inv %*% S %*% H_inv), "sandwich")
        return(if (is.null(out)) empty else out)
      }
      return(empty)
    }
    
    if (vcov_type == "auto") {
      if (!is.null(H_inv) && !is.null(S) && all(is.finite(H_inv)) && all(is.finite(S))) {
        out <- make_result(.clean_sym_matrix(H_inv %*% S %*% H_inv), "sandwich")
        if (!is.null(out)) {
          return(out)
        }
      }
      
      if (!is.null(S) && all(is.finite(S))) {
        S_inv <- .clean_sym_matrix(.safe_invert(S))
        out <- make_result(S_inv, "opg")
        if (!is.null(out)) {
          return(out)
        }
      }
    }
  }
  
  empty
}

.restricted_lambda_delta_se <- function(alpha, beta, gamma, vcov_sub, rho = 1) {
  alpha_sum <- sum(alpha)
  beta_sum <- sum(beta)
  den <- 1 - rho * (alpha_sum + beta_sum)
  
  if (!is.finite(den) || den <= 0) {
    return(NA_real_)
  }
  
  grad <- c(
    gamma = rho * alpha_sum / den,
    stats::setNames(
      rep(rho * gamma * (1 - rho * beta_sum) / (den^2), length(alpha)),
      names(alpha)
    ),
    stats::setNames(
      rep(rho^2 * gamma * alpha_sum / (den^2), length(beta)),
      names(beta)
    )
  )
  
  grad <- grad[rownames(vcov_sub)]
  se2 <- as.numeric(t(grad) %*% vcov_sub %*% grad)
  
  if (!is.finite(se2) || se2 <= 0) {
    return(NA_real_)
  }
  
  sqrt(se2)
}

.deduplicate_start_list <- function(starts) {
  if (length(starts) <= 1L) {
    return(starts)
  }
  
  key <- vapply(
    starts,
    function(z) paste(signif(unname(z), 8), collapse = "|"),
    character(1)
  )
  
  starts[!duplicated(key)]
}

.prepare_start_list <- function(start, x, model, arch_order, garch_order) {
  defaults <- qgarch_default_starts(
    x,
    model = model,
    arch_order = arch_order,
    garch_order = garch_order
  )
  
  if (is.null(start)) {
    return(defaults)
  }
  
  user_starts <- NULL
  
  if (is.numeric(start)) {
    user_starts <- list(.make_named_start(model, arch_order, garch_order, start))
  } else if (is.list(start) && length(start) > 0L && all(vapply(start, is.numeric, logical(1)))) {
    user_starts <- lapply(start, function(z) .make_named_start(model, arch_order, garch_order, z))
  } else {
    stop("'start' must be NULL, a numeric vector, or a list of numeric vectors.", call. = FALSE)
  }
  
  .deduplicate_start_list(c(user_starts, defaults))
}

.is_better_fit <- function(candidate, incumbent) {
  if (is.null(incumbent)) {
    return(TRUE)
  }
  
  cand_min <- candidate$minimum
  inc_min <- incumbent$minimum
  
  if (cand_min < inc_min - 1e-8) {
    return(TRUE)
  }
  if (abs(cand_min - inc_min) <= 1e-8 && candidate$code < incumbent$code) {
    return(TRUE)
  }
  
  FALSE
}

.refine_fit <- function(fit_obj,
                        x,
                        model,
                        arch_order,
                        garch_order,
                        rho,
                        indicator,
                        steptol,
                        iterlim,
                        print.level,
                        hessian) {
  p0 <- unname(fit_obj$estimate)
  typsize2 <- pmax(abs(p0), 0.05)
  
  refit <- try(
    stats::nlm(
      f = .qgarch_nll,
      p = p0,
      x = x,
      model = model,
      arch_order = arch_order,
      garch_order = garch_order,
      rho = rho,
      indicator = indicator,
      hessian = hessian,
      steptol = steptol / 10,
      typsize = typsize2,
      iterlim = max(iterlim, 500L),
      print.level = print.level
    ),
    silent = TRUE
  )
  
  if (inherits(refit, "try-error") || !is.finite(refit$minimum) || refit$minimum >= .qgarch_big_penalty) {
    return(fit_obj)
  }
  
  names(refit$estimate) <- names(fit_obj$estimate)
  
  if (.is_better_fit(refit, fit_obj)) {
    return(refit)
  }
  
  fit_obj
}

.build_fit_object <- function(best_fit,
                              x,
                              model,
                              arch_order,
                              garch_order,
                              rho,
                              indicator,
                              starts,
                              call,
                              vcov_type) {
  details <- .qgarch_nll(
    best_fit$estimate,
    x = x,
    model = model,
    arch_order = arch_order,
    garch_order = garch_order,
    rho = rho,
    indicator = indicator,
    return_details = TRUE
  )
  
  est_names <- names(best_fit$estimate)
  
  vcov_info <- .compute_vcov(
    theta = best_fit$estimate,
    raw_hessian = best_fit$hessian,
    x = x,
    model = model,
    arch_order = arch_order,
    garch_order = garch_order,
    rho = rho,
    indicator = indicator,
    vcov_type = vcov_type
  )
  
  vcov_mat <- vcov_info$vcov
  se_vec <- vcov_info$se
  
  names(se_vec) <- est_names
  rownames(vcov_mat) <- est_names
  colnames(vcov_mat) <- est_names
  
  coef_full <- best_fit$estimate
  full_se <- se_vec
  
  if (model == "zero") {
    coef_full <- c(coef_full, lambda = 0)
    full_se <- c(full_se, lambda = NA_real_)
  }
  
  implied_lambda <- NULL
  implied_lambda_se <- NULL
  
  if (model == "restricted") {
    alpha_idx <- paste0("alpha", seq_len(arch_order))
    beta_idx <- paste0("beta", seq_len(garch_order))
    
    alpha <- best_fit$estimate[alpha_idx]
    beta <- best_fit$estimate[beta_idx]
    gamma <- best_fit$estimate["gamma"]
    
    implied_lambda <- .restricted_lambda_value(alpha, beta, gamma, rho = rho)
    
    needed <- c("gamma", alpha_idx, beta_idx)
    if (all(needed %in% rownames(vcov_mat))) {
      implied_lambda_se <- .restricted_lambda_delta_se(
        alpha = alpha,
        beta = beta,
        gamma = gamma,
        vcov_sub = vcov_mat[needed, needed, drop = FALSE],
        rho = rho
      )
    } else {
      implied_lambda_se <- NA_real_
    }
    
    coef_full <- c(coef_full, lambda = implied_lambda)
    full_se <- c(full_se, lambda = implied_lambda_se)
  }
  
  structure(
    list(
      call = call,
      model = model,
      arch_order = arch_order,
      garch_order = garch_order,
      coefficients = best_fit$estimate,
      coefficients_full = coef_full,
      standard_errors = se_vec,
      standard_errors_full = full_se,
      vcov = vcov_mat,
      vcov_type = vcov_info$method,
      loglik = -best_fit$minimum,
      nll = best_fit$minimum,
      eta = details$eta,
      sigma2 = details$sigma2,
      fitted = best_fit$estimate["mu"] + best_fit$estimate["gamma"] * details$sigma2,
      residuals = x - (best_fit$estimate["mu"] + best_fit$estimate["gamma"] * details$sigma2),
      standardized_residuals = details$eta / sqrt(details$sigma2),
      lambda_t = details$lambda,
      kappa = details$kappa,
      unconditional_variance = details$uncond_var,
      hessian = best_fit$hessian,
      convergence = best_fit$code,
      iterations = best_fit$iterations,
      minimum = best_fit$minimum,
      x = x,
      indicator = indicator,
      rho = rho,
      starts_tried = starts,
      implied_lambda = implied_lambda,
      implied_lambda_se = implied_lambda_se
    ),
    class = "qgarch"
  )
}

#' Fit generalized qgarch(m, n) models
#'
#' Fits QGARCH-in-mean models using nonlinear minimization of the negative
#' log-likelihood. Four variants are supported: a zero-lambda model, a
#' restricted-lambda model, a free-lambda model, and a threshold model with
#' state-dependent lambda.
#'
#' @param x A numeric vector or one-column data frame containing the observed
#'   series. The series must contain only finite, non-missing values.
#' @param model Character string specifying the model variant. Must be one of
#'   `"zero"`, `"restricted"`, `"free"`, or `"threshold"`.
#' @param arch_order Positive integer giving the ARCH lag order `m`.
#' @param garch_order Positive integer giving the GARCH lag order `n`.
#' @param threshold_indicator Optional threshold indicator used only when
#'   `model = "threshold"`. Must have the same length as `x`. Nonzero values
#'   are converted to 1 and zero values to 0.
#' @param start Optional starting values. May be `NULL`, a numeric vector, or a
#'   list of numeric vectors.
#' @param rho Scalar used in the restricted-lambda mapping.
#' @param steptol Step tolerance passed to [stats::nlm()].
#' @param typsize Typical size passed to [stats::nlm()]. A single value is
#'   repeated to the appropriate length.
#' @param iterlim Maximum number of iterations passed to [stats::nlm()].
#' @param print.level Print level passed to [stats::nlm()].
#' @param hessian Logical; should the Hessian be returned by [stats::nlm()]?
#' @param vcov_type Character string controlling standard-error estimation.
#'   One of `"auto"`, `"sandwich"`, `"hessian"`, `"opg"`, or `"none"`.
#'   In `"auto"` mode, Hessian-based standard errors are tried first.
#'
#' @return An object of class `"qgarch"`.
#' @export
qgarch_fit <- function(x,
                       model = c("zero", "restricted", "free", "threshold"),
                       arch_order = 1L,
                       garch_order = 1L,
                       threshold_indicator = NULL,
                       start = NULL,
                       rho = 1,
                       steptol = 1e-10,
                       typsize = 0.1,
                       iterlim = 300L,
                       print.level = 0,
                       hessian = TRUE,
                       vcov_type = c("auto", "sandwich", "hessian", "opg", "none")) {
  model <- match.arg(model)
  vcov_type <- match.arg(vcov_type)
  
  orders <- .validate_orders(arch_order, garch_order)
  arch_order <- orders$arch_order
  garch_order <- orders$garch_order
  
  x <- .validate_input_series(x)
  
  if (!is.numeric(rho) || length(rho) != 1L || !is.finite(rho)) {
    stop("'rho' must be a single finite numeric value.", call. = FALSE)
  }
  
  if (!is.numeric(typsize) || length(typsize) != 1L || !is.finite(typsize) || typsize <= 0) {
    stop("'typsize' must be a single positive finite numeric value.", call. = FALSE)
  }
  
  if (!is.numeric(steptol) || length(steptol) != 1L || !is.finite(steptol) || steptol <= 0) {
    stop("'steptol' must be a single positive finite numeric value.", call. = FALSE)
  }
  
  if (!is.numeric(iterlim) || length(iterlim) != 1L || !is.finite(iterlim) || iterlim < 1) {
    stop("'iterlim' must be a single positive integer.", call. = FALSE)
  }
  iterlim <- as.integer(iterlim)
  
  if (!is.numeric(print.level) || length(print.level) != 1L || !is.finite(print.level)) {
    stop("'print.level' must be a single finite numeric value.", call. = FALSE)
  }
  
  if (!is.logical(hessian) || length(hessian) != 1L || is.na(hessian)) {
    stop("'hessian' must be either TRUE or FALSE.", call. = FALSE)
  }
  
  if (model == "threshold") {
    if (is.null(threshold_indicator)) {
      stop("'threshold_indicator' must be supplied when model = 'threshold'.", call. = FALSE)
    }
    
    indicator <- as.numeric(threshold_indicator)
    
    if (length(indicator) != length(x)) {
      stop("'threshold_indicator' must have the same length as 'x'.", call. = FALSE)
    }
    
    if (anyNA(indicator) || any(!is.finite(indicator))) {
      stop("'threshold_indicator' must contain only finite, non-missing values.", call. = FALSE)
    }
    
    indicator <- ifelse(indicator != 0, 1, 0)
  } else {
    indicator <- NULL
  }
  
  starts <- .prepare_start_list(
    start = start,
    x = x,
    model = model,
    arch_order = arch_order,
    garch_order = garch_order
  )
  
  best_fit <- NULL
  
  for (i in seq_along(starts)) {
    fit_i <- try(
      stats::nlm(
        f = .qgarch_nll,
        p = unname(starts[[i]]),
        x = x,
        model = model,
        arch_order = arch_order,
        garch_order = garch_order,
        rho = rho,
        indicator = indicator,
        hessian = hessian,
        steptol = steptol,
        typsize = rep(typsize, length(starts[[i]])),
        iterlim = iterlim,
        print.level = print.level
      ),
      silent = TRUE
    )
    
    if (!inherits(fit_i, "try-error") &&
        is.finite(fit_i$minimum) &&
        fit_i$minimum < .qgarch_big_penalty) {
      names(fit_i$estimate) <- names(starts[[i]])
      if (.is_better_fit(fit_i, best_fit)) {
        best_fit <- fit_i
      }
    }
  }
  
  if (is.null(best_fit)) {
    stop("No converged solution was found. Try supplying custom starting values.", call. = FALSE)
  }
  
  best_fit <- .refine_fit(
    fit_obj = best_fit,
    x = x,
    model = model,
    arch_order = arch_order,
    garch_order = garch_order,
    rho = rho,
    indicator = indicator,
    steptol = steptol,
    iterlim = iterlim,
    print.level = print.level,
    hessian = hessian
  )
  
  .build_fit_object(
    best_fit = best_fit,
    x = x,
    model = model,
    arch_order = arch_order,
    garch_order = garch_order,
    rho = rho,
    indicator = indicator,
    starts = starts,
    call = match.call(),
    vcov_type = vcov_type
  )
}

# Replication notes

This package was built from the uploaded Campbell and Hentschel (1992) paper
and the supplied QGARCH replication scripts.

Implemented directly from the supplied code:

- QGARCH-M(1,1) with `lambda = 0`
- QGARCH-M(1,1) with restricted `lambda`
- QGARCH-M(1,1) with free `lambda`
- Threshold extension with `lambda_t = lambda1 + lambda2 * I_t`
- Hessian-based standard errors with pseudoinverse fallback
- Likelihood-ratio tests between nested specifications
- Table 1 style moment diagnostics

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{coef.qgarch}
\alias{coef.qgarch}
\title{Extract coefficients from a qgarch model}
\usage{
\method{coef}{qgarch}(object, ...)
}
\arguments{
\item{object}{A fitted \code{qgarch} object.}

\item{...}{Additional arguments. Supports \code{type = "estimated"} or
\code{type = "full"}.}
}
\value{
A named numeric vector of coefficients.
}
\description{
Extract coefficients from a qgarch model
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{predict.qgarch}
\alias{predict.qgarch}
\title{Forecast from a generalized qgarch(m, n) model}
\usage{
\method{predict}{qgarch}(object, n.ahead = 1L, ...)
}
\arguments{
\item{object}{A fitted \code{qgarch} object.}

\item{n.ahead}{Number of periods ahead to forecast.}

\item{...}{Unused.}
}
\value{
A data frame with forecast horizon, conditional mean,
conditional variance, and conditional standard deviation.
}
\description{
Forecast from a generalized qgarch(m, n) model
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{summary.qgarch}
\alias{summary.qgarch}
\title{Summarize a qgarch model}
\usage{
\method{summary}{qgarch}(object, ...)
}
\arguments{
\item{object}{A fitted \code{qgarch} object.}

\item{...}{Unused.}
}
\value{
An object of class \code{"summary.qgarch"}.
}
\description{
Summarize a qgarch model
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/qgarch_fit.R
\name{qgarch_default_starts}
\alias{qgarch_default_starts}
\title{Default starting values for qgarch estimation}
\usage{
qgarch_default_starts(
  x,
  model = c("zero", "restricted", "free", "threshold"),
  arch_order = 1L,
  garch_order = 1L
)
}
\arguments{
\item{x}{A numeric vector or one-column data frame containing the observed
series.}

\item{model}{Character string specifying the model variant. Must be one of
\code{"zero"}, \code{"restricted"}, \code{"free"}, or \code{"threshold"}.}

\item{arch_order}{Positive integer giving the ARCH lag order.}

\item{garch_order}{Positive integer giving the GARCH lag order.}
}
\value{
A list of named numeric vectors containing candidate starting values.
}
\description{
Creates a compact but order-adaptive set of starting values for the selected
qgarch model.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{print.qgarch}
\alias{print.qgarch}
\title{Print a qgarch model}
\usage{
\method{print}{qgarch}(x, digits = max(3L, getOption("digits") - 2L), ...)
}
\arguments{
\item{x}{A fitted \code{qgarch} object.}

\item{digits}{Number of digits to print.}

\item{...}{Unused.}
}
\value{
The input object, invisibly.
}
\description{
Print a qgarch model
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{qgarch_lr_test}
\alias{qgarch_lr_test}
\title{Likelihood ratio test for nested qgarch models}
\usage{
qgarch_lr_test(restricted, unrestricted, df = NULL)
}
\arguments{
\item{restricted}{A fitted \code{qgarch} object representing the restricted model.}

\item{unrestricted}{A fitted \code{qgarch} object representing the unrestricted model.}

\item{df}{Degrees of freedom for the test. If \code{NULL}, it is computed as the
difference in the number of estimated coefficients.}
}
\value{
A data frame with the likelihood ratio statistic, degrees of freedom,
p-value, and the model names. The returned object has class
\code{c("qgarch_lr_test", "data.frame")}.
}
\description{
Compares two fitted \code{qgarch} models using the likelihood ratio test.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{us_monthly}
\alias{us_monthly}
\title{Monthly U.S. data}
\format{
A data frame with monthly observations and 3 variables:
\describe{
\item{YearMon}{Month of the observation.}
\item{Mktrf}{Market excess return.}
\item{ER}{Excess return series.}
}
}
\source{
Prepared by the package author.
}
\usage{
us_monthly
}
\description{
Monthly data used in examples and tests for the qgarch package.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{logLik.qgarch}
\alias{logLik.qgarch}
\title{Log-likelihood for a qgarch model}
\usage{
\method{logLik}{qgarch}(object, ...)
}
\arguments{
\item{object}{A fitted \code{qgarch} object.}

\item{...}{Unused.}
}
\value{
An object of class \code{"logLik"}.
}
\description{
Log-likelihood for a qgarch model
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{qgarch_moments}
\alias{qgarch_moments}
\title{Compute sample moments for a series}
\usage{
qgarch_moments(x, scale_mean_variance = 1)
}
\arguments{
\item{x}{A numeric vector or a one-column data frame containing the series.}

\item{scale_mean_variance}{A finite scalar used to rescale the reported mean
and variance and their standard errors.}
}
\value{
A list with elements:
\describe{
\item{estimate}{Named numeric vector of moment estimates.}
\item{std_error}{Named numeric vector of standard errors.}
\item{n}{Number of usable observations.}
}
The returned object has class \code{"qgarch_moments"}.
}
\description{
Computes the sample mean, variance, skewness, and excess kurtosis
for a numeric series, along with simple standard errors.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{vcov.qgarch}
\alias{vcov.qgarch}
\title{Variance-covariance matrix for a qgarch model}
\usage{
\method{vcov}{qgarch}(object, ...)
}
\arguments{
\item{object}{A fitted \code{qgarch} object.}

\item{...}{Unused.}
}
\value{
A variance-covariance matrix.
}
\description{
Variance-covariance matrix for a qgarch model
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/qgarch_fit.R
\name{qgarch_fit}
\alias{qgarch_fit}
\title{Fit generalized qgarch(m, n) models}
\usage{
qgarch_fit(
  x,
  model = c("zero", "restricted", "free", "threshold"),
  arch_order = 1L,
  garch_order = 1L,
  threshold_indicator = NULL,
  start = NULL,
  rho = 1,
  steptol = 1e-10,
  typsize = 0.1,
  iterlim = 300L,
  print.level = 0,
  hessian = TRUE,
  vcov_type = c("auto", "sandwich", "hessian", "opg", "none")
)
}
\arguments{
\item{x}{A numeric vector or one-column data frame containing the observed
series. The series must contain only finite, non-missing values.}

\item{model}{Character string specifying the model variant. Must be one of
\code{"zero"}, \code{"restricted"}, \code{"free"}, or \code{"threshold"}.}

\item{arch_order}{Positive integer giving the ARCH lag order \code{m}.}

\item{garch_order}{Positive integer giving the GARCH lag order \code{n}.}

\item{threshold_indicator}{Optional threshold indicator used only when
\code{model = "threshold"}. Must have the same length as \code{x}. Nonzero values
are converted to 1 and zero values to 0.}

\item{start}{Optional starting values. May be \code{NULL}, a numeric vector, or a
list of numeric vectors.}

\item{rho}{Scalar used in the restricted-lambda mapping.}

\item{steptol}{Step tolerance passed to \code{\link[stats:nlm]{stats::nlm()}}.}

\item{typsize}{Typical size passed to \code{\link[stats:nlm]{stats::nlm()}}. A single value is
repeated to the appropriate length.}

\item{iterlim}{Maximum number of iterations passed to \code{\link[stats:nlm]{stats::nlm()}}.}

\item{print.level}{Print level passed to \code{\link[stats:nlm]{stats::nlm()}}.}

\item{hessian}{Logical; should the Hessian be returned by \code{\link[stats:nlm]{stats::nlm()}}?}

\item{vcov_type}{Character string controlling standard-error estimation.
One of \code{"auto"}, \code{"sandwich"}, \code{"hessian"}, \code{"opg"}, or \code{"none"}.
In \code{"auto"} mode, Hessian-based standard errors are tried first.}
}
\value{
An object of class \code{"qgarch"}.
}
\description{
Fits QGARCH-in-mean models using nonlinear minimization of the negative
log-likelihood. Four variants are supported: a zero-lambda model, a
restricted-lambda model, a free-lambda model, and a threshold model with
state-dependent lambda.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{fitted.qgarch}
\alias{fitted.qgarch}
\title{Fitted values from a qgarch model}
\usage{
\method{fitted}{qgarch}(object, ...)
}
\arguments{
\item{object}{A fitted \code{qgarch} object.}

\item{...}{Unused.}
}
\value{
A numeric vector of fitted values.
}
\description{
Fitted values from a qgarch model
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{residuals.qgarch}
\alias{residuals.qgarch}
\title{Residuals from a qgarch model}
\usage{
\method{residuals}{qgarch}(object, type = c("raw", "standardized", "eta"), ...)
}
\arguments{
\item{object}{A fitted \code{qgarch} object.}

\item{type}{Type of residuals to return: \code{"raw"}, \code{"standardized"},
or \code{"eta"}.}

\item{...}{Unused.}
}
\value{
A numeric vector of residuals.
}
\description{
Residuals from a qgarch model
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{plot.qgarch}
\alias{plot.qgarch}
\title{Plot a qgarch model}
\usage{
\method{plot}{qgarch}(x, which = c("sigma2", "standardized"), ...)
}
\arguments{
\item{x}{A fitted \code{qgarch} object.}

\item{which}{Which plot to show: \code{"sigma2"} or \code{"standardized"}.}

\item{...}{Additional graphical arguments passed to \code{\link[graphics:plot]{graphics::plot()}}.}
}
\value{
The input object, invisibly.
}
\description{
Plot a qgarch model
}
